from celery import Celery

from .celery import patch_celery
from .notifier import Notifier

app = Celery("test_celery", broker="redis://localhost", backend="redis://localhost")

notifier = Notifier(
    host="http://localhost:8080", environment="celery", apm_disabled=True
)
patch_celery(notifier)


@app.task
def raise_error():
    raise ValueError("Test")



# API version to poll.
_API_VER = "2020-06-18"

# How frequently we should poll the config API.
_DEFAULT_INTERVAL = 600

# What path to poll.
_CONFIG_ROUTE_PATTERN = "%s/%s/config/%d/config.json"

# Remote config settings.
_ERROR_SETTING = "errors"
_APM_SETTING = "apm"


class SettingsData:
    def __init__(self, project_id, data):
        self._project_id = project_id
        self._data = data

    def merge(self, other_data):
        self._data = {**self._data, **other_data}

    def interval(self):
        poll_sec = self._data.get("poll_sec")
        if poll_sec is None:
            return _DEFAULT_INTERVAL

        if poll_sec > 0:
            return poll_sec

        return _DEFAULT_INTERVAL

    def config_route(self, remote_config_host):
        host = remote_config_host.rstrip("/")

        config_route = self._data.get("config_route")
        if config_route:
            return host + "/" + config_route

        return _CONFIG_ROUTE_PATTERN % (host, _API_VER, self._project_id)

    def error_notifications(self):
        s = self._find_setting(_ERROR_SETTING)
        if s is None:
            return True

        return s.get("enabled")

    def performance_stats(self):
        s = self._find_setting(_APM_SETTING)
        if s is None:
            return True

        return s.get("enabled")

    def error_host(self):
        s = self._find_setting(_ERROR_SETTING)
        if s is None:
            return None

        return s.get("endpoint")

    def apm_host(self):
        s = self._find_setting(_APM_SETTING)
        if s is None:
            return None

        return s.get("endpoint")

    def _find_setting(self, name):
        settings = self._data.get("settings")
        if settings is None:
            return None

        for i in settings:
            if i.get("name") == name:
                return i

        return None



import base64
import struct
import tdigest


class TDigestStat:
    __slots__ = ("count", "sum", "sumsq", "td", "tdigest")

    def __init__(self):
        self.count = 0
        self.sum = 0
        self.sumsq = 0
        self.td = tdigest.TDigest(K=10)
        self.tdigest = None

    @property
    def __dict__(self):
        b = as_bytes(self.td)
        self.tdigest = base64.b64encode(b).decode("ascii")
        return {s: getattr(self, s) for s in self.__slots__ if s != "td"}

    def add(self, ms):
        self.count += 1
        self.sum += ms
        self.sumsq += ms * ms
        self.td.update(ms)


class TDigestStatGroups(TDigestStat):
    __slots__ = TDigestStat.__slots__ + ("groups",)

    def __init__(self):
        super().__init__()
        self.groups = {}

    def add_groups(self, total_ms, groups):
        self.add(total_ms)

        for name, ms in groups.items():
            self.add_group(name, ms)

    def add_group(self, name, ms):
        stat = self.groups.get(name)
        if stat is None:
            stat = TDigestStat()
            self.groups[name] = stat
        stat.add(ms)


_SMALL_ENCODING = 2


def from_bytes(b):
    encoding = int.from_bytes(b[:4], byteorder="big")
    if encoding != _SMALL_ENCODING:
        raise ValueError(f"unsupported encoding version: {encoding}")

    compression = struct.unpack(">d", b[4:12])[0]
    num_centroids = int.from_bytes(b[12:16], byteorder="big")
    if num_centroids < 0 or num_centroids > 1 << 22:
        raise ValueError("bad number of centroids")

    b = b[16:]
    digest = tdigest.TDigest(K=int(compression))

    x = 0
    means = []
    for i in range(num_centroids):
        delta = struct.unpack(">f", b[:4])[0]
        b = b[4:]
        x += delta
        means.append(x)

    stream = decode_varint_stream(b)
    for i in range(num_centroids):
        count = next(stream)
        digest.update(means[i], count)

    return digest


def as_bytes(digest):
    b = bytearray()

    b += _SMALL_ENCODING.to_bytes(4, byteorder="big")
    b += struct.pack(">d", digest.K)
    b += len(digest).to_bytes(4, byteorder="big")

    x = 0
    for key in digest.C.keys():
        delta = key - x
        x = key
        b += struct.pack(">f", delta)

    for value in digest.C.values():
        encode_varint(b, int(value.count))

    return b


def encode_varint(b, num):
    while True:
        c = num & 0x7F
        num >>= 7
        if num:
            b += bytes([c | 0x80])
        else:
            b += bytes([c])
            break


def decode_varint_stream(b):
    value = 0
    base = 1
    for c in b:
        value += (c & 0x7F) * base
        if c & 0x80:
            base <<= 7
        else:
            yield value
            value = 0
            base = 1



import base64
import json
import threading
import urllib.request
import urllib.error

from . import metrics
from .tdigest import as_bytes, TDigestStatGroups
from .utils import logger, time_trunc_minute


class _RouteBreakdown(TDigestStatGroups):
    __slots__ = TDigestStatGroups.__slots__ + (
        "method",
        "route",
        "responseType",
        "time",
    )

    def __init__(self, *, method="", route="", responseType="", time=None):
        super().__init__()
        self.method = method
        self.route = route
        self.responseType = responseType
        self.time = time_trunc_minute(time)

    @property
    def __dict__(self):
        b = as_bytes(self.td)
        self.tdigest = base64.b64encode(b).decode("ascii")
        d = {s: getattr(self, s) for s in self.__slots__ if s != "td"}

        groups = d["groups"]
        for k, v in groups.items():
            groups[k] = v.__dict__

        return d


class RouteBreakdowns:
    def __init__(self, *, project_id=0, project_key="", **kwargs):
        self._config = kwargs["config"]

        self._project_id = project_id
        self._ab_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + project_key,
        }
        self._env = kwargs.get("environment")

        self._thread = None
        self._lock = threading.Lock()
        self._stats = None

    def notify(self, metric):
        if not self._config.get("performance_stats"):
            return

        if (
            metric.status_code < 200
            or (metric.status_code >= 300 and metric.status_code < 400)
            or metric.status_code == 404
            or len(metric._groups) <= 1
        ):
            return

        if self._stats is None:
            self._stats = {}
            self._thread = threading.Timer(metrics.FLUSH_PERIOD, self._flush)
            self._thread.start()

        key = metric._key()
        with self._lock:
            if key in self._stats:
                stat = self._stats[key]
            else:
                stat = _RouteBreakdown(
                    method=metric.method,
                    route=metric.route,
                    responseType=metric.response_type,
                    time=metric.start_time,
                )
                self._stats[key] = stat

            total_ms = (metric.end_time - metric.start_time) * 1000
            stat.add_groups(total_ms, metric._groups)

    def _flush(self):
        stats = None
        with self._lock:
            stats = self._stats
            self._stats = None

        if not stats:
            raise ValueError("stats is empty")

        out = {"routes": [v.__dict__ for v in stats.values()]}
        if self._env:
            out["environment"] = self._env

        out_json = json.dumps(out).encode("utf8")
        req = urllib.request.Request(
            self._ab_url(), data=out_json, headers=self._ab_headers, method="POST"
        )

        try:
            resp = urllib.request.urlopen(req, timeout=5)
        except urllib.error.HTTPError as err:
            resp = err
        except Exception as err:  # pylint: disable=broad-except
            logger.error(err)
            return

        try:
            body = resp.read()
        except IOError as err:
            logger.error(err)
            return

        if 200 <= resp.code < 300:
            return

        if not 400 <= resp.code < 500:
            err = f"airbrake: unexpected response status_code={resp.code}"
            logger.error(err)
            return

        if resp.code == 429:
            return

        try:
            body = body.decode("utf-8")
        except UnicodeDecodeError as err:
            logger.error(err)
            return

        try:
            in_data = json.loads(body)
        except ValueError as err:  # json.JSONDecodeError requires Python 3.5+
            logger.error(err)
            return

        if "message" in in_data:
            logger.error(in_data["message"])
            return

    def _ab_url(self):
        return f"{self._config.get('apm_host')}/api/v5/projects/" \
               f"{self._project_id}/routes-breakdowns"


HTTP_HANDLER = "http.handler"


class RouteMetric(metrics.Metric):
    def __init__(self, *, method="", route="", status_code=0, content_type=""):
        super().__init__()
        self.method = method
        self.route = route
        self.status_code = status_code
        self.content_type = content_type
        self.start_span(HTTP_HANDLER, start_time=self.start_time)

    def end(self):
        super().end()
        self.end_span(HTTP_HANDLER, end_time=self.end_time)

    def _key(self):
        time = self.start_time // 60 * 60
        return (self.method, self.route, self.response_type, time)

    @property
    def response_type(self):
        if self.status_code >= 500:
            return "5xx"
        if self.status_code >= 400:
            return "4xx"
        return self.content_type.split(";")[0].split("/")[-1]



import base64
import json
import threading
import urllib.request
import urllib.error

from . import metrics
from .tdigest import TDigestStat, as_bytes
from .utils import logger, time_trunc_minute


class QueryStat(TDigestStat):
    __slots__ = TDigestStat.__slots__ + ("query", "method", "route", "time")

    @property
    def __dict__(self):
        tdigest = as_bytes(self.td)
        self.tdigest = base64.b64encode(tdigest).decode("ascii")

        return {s: getattr(self, s) for s in self.__slots__ if s != "td"}

    def __init__(self, *, query="", method="", route="", time=None):
        super().__init__()
        self.query = query
        self.method = method
        self.route = route
        self.time = time_trunc_minute(time)


class QueryStats:
    def __init__(self, *, project_id=0, project_key="", **kwargs):
        self._config = kwargs["config"]

        self._project_id = project_id
        self._ab_headers = {
            "Content-Type": "application/json",
            "Authorization": "Bearer " + project_key,
        }
        self._env = kwargs.get("environment")

        self._thread = None
        self._lock = threading.Lock()
        self._stats = None

    def notify(self, *, query="", method="", route="", start_time=None, end_time=None):
        if not self._config.get("performance_stats"):
            return
        if not self._config.get("query_stats"):
            return

        key = query_stat_key(query=query, method=method, route=route, time=start_time)
        ms = (end_time - start_time) * 1000

        with self._lock:
            if self._stats is None:
                self._stats = {}
                self._thread = threading.Timer(metrics.FLUSH_PERIOD, self._flush)
                self._thread.start()

            if key in self._stats:
                stat = self._stats[key]
            else:
                stat = QueryStat(
                    query=query, method=method, route=route, time=start_time
                )
                self._stats[key] = stat
            stat.add(ms)

    def _flush(self):
        stats = None
        with self._lock:
            stats = self._stats
            self._stats = None

        if not stats:
            raise ValueError("stats is empty")

        out = {"queries": [v.__dict__ for v in stats.values()]}
        if self._env:
            out["environment"] = self._env

        out = json.dumps(out).encode("utf8")
        req = urllib.request.Request(
            self._ab_url(), data=out, headers=self._ab_headers, method="POST"
        )

        try:
            resp = urllib.request.urlopen(req, timeout=5)
        except urllib.error.HTTPError as err:
            resp = err
        except Exception as err:  # pylint: disable=broad-except
            logger.error(err)
            return

        try:
            body = resp.read()
        except IOError as err:
            logger.error(err)
            return

        if 200 <= resp.code < 300:
            return

        if not 400 <= resp.code < 500:
            err = f"airbrake: unexpected response status_code={resp.code}"
            logger.error(err)
            return

        if resp.code == 429:
            return

        try:
            body = body.decode("utf-8")
        except UnicodeDecodeError as err:
            logger.error(err)
            return

        try:
            in_data = json.loads(body)
        except ValueError as err:  # json.JSONDecodeError requires Python 3.5+
            logger.error(err)
            return

        if "message" in in_data:
            logger.error(in_data["message"])
            return

    def _ab_url(self):
        return f"{self._config.get('apm_host')}/api/v5/projects/{self._project_id}/queries-stats"


def query_stat_key(*, query="", method="", route="", time=None):
    time = time // 60 * 60
    return (query, method, route, time)



import time
from flask import (
    request,
    got_request_exception,
    before_render_template,
    template_rendered,
)

try:
    import flask_sqlalchemy as _
except ImportError:
    _sqla_available = False
else:
    _sqla_available = True

try:
    from flask_login import current_user
except ImportError:
    _flask_login_available = False
else:
    _flask_login_available = True

from .notifier import Notifier
from .route_metric import RouteMetric
from . import metrics


_UNKNOWN_ROUTE = "UNKNOWN"


def request_filter(notice):
    if request is None:
        return notice

    ctx = notice["context"]
    ctx["method"] = request.method
    ctx["url"] = request.url
    ctx["route"] = str(request.endpoint)

    try:
        user_addr = request.access_route[0]
    except IndexError:
        user_addr = request.remote_addr
    if user_addr:
        ctx["userAddr"] = user_addr

    if _flask_login_available and current_user.is_authenticated:
        user = dict(id=current_user.get_id())
        for s in ["username", "name"]:
            if hasattr(current_user, s):
                user[s] = getattr(current_user, s)
        ctx["user"] = user

    notice["params"]["request"] = dict(
        form=dict(request.form),
        json=dict(request.json),
        files=dict(request.files),
        cookies=dict(request.cookies),
        headers=dict(request.headers),
        environ=request.environ,
        blueprint=request.blueprint,
        url_rule=request.url_rule,
        view_args=request.view_args,
    )

    return notice


def init_app(app):
    if "pybrake" in app.extensions:
        raise ValueError("pybrake is already injected")
    if "PYBRAKE" not in app.config:
        raise ValueError("app.config['PYBRAKE'] is not defined")

    notifier = Notifier(**app.config["PYBRAKE"])
    config = notifier.config

    notifier.add_filter(request_filter)

    app.extensions["pybrake"] = notifier
    got_request_exception.connect(_handle_exception, sender=app)

    before_render_template.connect(_before_render_template, sender=app)
    template_rendered.connect(_template_rendered, sender=app)

    app.before_request(_before_request(config, notifier))
    app.after_request(_after_request(config, notifier))

    if _sqla_available:
        _sqla_instrument(config, app)

    return app


def _before_request(config, notifier):
    def before_request_middleware():
        if not config.get("performance_stats"):
            return

        if request.url_rule:
            route = request.url_rule.rule
        else:
            route = _UNKNOWN_ROUTE

        metric = RouteMetric(method=request.method, route=route)
        metrics.set_active(metric)

    return before_request_middleware


def _after_request(config, notifier):
    def after_request_middleware(response):
        if not config.get("performance_stats"):
            return response

        metric = metrics.get_active()
        if metric is not None:
            metric.status_code = response.status_code
            metric.content_type = response.headers.get("Content-Type")
            metric.end_time = time.time()
            notifier.routes.notify(metric)
            metrics.set_active(None)

        return response

    return after_request_middleware


def _before_render_template(sender, template, context, **extra):
    metrics.start_span("template")


def _template_rendered(sender, template, context, **extra):
    metrics.end_span("template")


def _handle_exception(sender, exception, **_):
    notifier = sender.extensions["pybrake"]

    notice = notifier.build_notice(exception)
    notifier.send_notice(notice)


def _sqla_instrument(config, app):
    try:
        sqla = app.extensions["sqlalchemy"]
    except Exception:  # pylint: disable=broad-except
        return

    from sqlalchemy import event  # pylint: disable=import-outside-toplevel

    engine = sqla.db.get_engine()
    event.listen(engine, "before_cursor_execute", _before_cursor(config))
    event.listen(engine, "after_cursor_execute", _after_cursor(config))

def _before_cursor(config):
    def _sqla_before_cursor_execute(
        conn, cursor, statement, parameters, context, executemany
    ):
        if not config.get("performance_stats"):
            return
        metrics.start_span("sql")

    return _sqla_before_cursor_execute

def _after_cursor(config):
    def _sqla_after_cursor_execute(
        conn, cursor, statement, parameters, context, executemany
    ):
        if not config.get("performance_stats"):
            return
        metrics.end_span("sql")

    return _sqla_after_cursor_execute



import re
import warnings
from urllib.error import URLError

from .notifier import Notifier
from .utils import time_trunc_minute
from .test_helper import (get_exception, get_nested_exception,
                          get_exception_in_cython, build_notice_from_str)


def test_build_notice_from_exception():
    notifier = Notifier()

    err = get_exception()
    notice = notifier.build_notice(err)

    errors = notice["errors"]
    assert len(errors) == 1

    error = errors[0]
    assert error["type"] == "ValueError"
    assert error["message"] == "hello"

    backtrace = error["backtrace"]
    assert len(backtrace) == 1

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_helper.py"
    assert frame["function"] == "get_exception"
    assert frame["line"] == 6
    assert frame["code"] == {
        4: "def get_exception():",
        5: "    try:",
        6: '        raise ValueError("hello")',
        7: "    except ValueError as err:",
        8: "        return err",
    }

    context = notice["context"]
    assert context["notifier"]["name"] == "pybrake"
    assert context["notifier"]["url"] == "https://github.com/airbrake/pybrake"
    assert context["notifier"]["version"]

    for k in ["os", "language", "hostname", "versions", "revision"]:
        assert context[k]


def test_build_notice_from_exception_cython():
    notifier = Notifier()

    err = get_exception_in_cython()
    notice = notifier.build_notice(err)

    errors = notice["errors"]
    assert len(errors) == 1

    error = errors[0]
    assert error["type"] == "TypeError"

    message = error["message"]
    if message.startswith('unorderable'):
        # python 3.5
        assert error["message"] == 'unorderable types: str() < int()'
    else:
        # python 3.6 and above
        assert error["message"] == "'<' not supported between instances of 'str' and 'int'"


def test_build_notice_from_nested_exception():
    notifier = Notifier()

    err = get_nested_exception()
    notice = notifier.build_notice(err)

    errors = notice["errors"]
    assert len(errors) == 2

    error = errors[0]
    assert error["type"] == "ValueError"
    assert error["message"] == "world"

    backtrace = error["backtrace"]
    assert len(backtrace) == 1

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_helper.py"
    assert frame["function"] == "get_nested_exception"
    assert frame["line"] == 51

    error = errors[1]
    assert error["type"] == "ValueError"
    assert error["message"] == "hello"

    backtrace = error["backtrace"]
    assert len(backtrace) == 2

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_helper.py"
    assert frame["function"] == "get_exception"
    assert frame["line"] == 6


def test_build_notice_from_str():
    notifier = Notifier()

    notice = build_notice_from_str(notifier, "hello")

    errors = notice["errors"]
    assert len(errors) == 1

    error = errors[0]
    assert error["message"] == "hello"

    backtrace = error["backtrace"]
    assert len(backtrace) >= 1

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_helper.py"
    assert frame["function"] == "build_notice_from_str"
    assert frame["line"] == 12
    assert frame["code"] == {
        10: "",
        11: "def build_notice_from_str(notifier, s):",
        12: "    return notifier.build_notice(s)",
        13: "",
        14: "",
    }


def test_build_notice_from_none():
    notifier = Notifier()
    notice = notifier.build_notice(None)

    errors = notice["errors"]
    assert not errors


def test_environment():
    notifier = Notifier(environment="production")

    notice = notifier.build_notice("hello")

    assert notice["context"]["environment"] == "production"


def test_root_directory():
    notifier = Notifier(root_directory="/root/dir")

    notice = notifier.build_notice("hello")

    assert notice["context"]["rootDirectory"] == "/root/dir"


def test_filter_data():
    def notifier_filter(notice):
        notice["params"]["param"] = "value"
        return notice

    notifier = Notifier()
    notifier.add_filter(notifier_filter)

    notice = notifier.notify_sync("hello")

    assert notice["params"]["param"] == "value"


def test_filter_ignore():
    notifier = Notifier()
    notifier.add_filter(lambda notice: None)

    notice = notifier.notify_sync("hello")

    assert notice["error"] == "notice is filtered out"


def test_filter_ignore_async():
    notifier = Notifier()
    notifier.add_filter(lambda notice: None)

    future = notifier.notify("hello")
    notice = future.result()

    assert notice["error"] == "notice is filtered out"


def test_pybrake_error_filter():
    notifier = Notifier()

    try:
        time_trunc_minute(None)
    except Exception as err:  # pylint: disable=broad-except
        notice = notifier.notify_sync(err)
        assert notice["error"] == "notice is filtered out"


def test_unauthorized():
    notifier = Notifier()

    notice = notifier.notify_sync("hello")

    assert notice["error"] == "Project API key is required"


def test_unknown_host():
    notifier = Notifier(host="http://airbrake123.com")

    notice = notifier.notify_sync("hello")

    assert isinstance(notice["error"], URLError)
    assert "not known>" in str(notice["error"])


def test_truncation():
    notifier = Notifier()

    notice = notifier.build_notice("hello")
    notice["params"]["param"] = "x" * 64000
    notice = notifier.send_notice_sync(notice)

    assert len(notice["params"]["param"]) == 1024


def test_revision_override():
    notifier = Notifier(revision="1234")
    assert notifier._context["revision"] == "1234"


def test_revision_from_git(monkeypatch):
    monkeypatch.setattr("pybrake.notifier.get_git_revision", lambda x: "4321")
    notifier = Notifier()
    assert notifier._context["revision"] == "4321"


def _test_keys_blocklist(keys_blocklist):
    notifier = Notifier(keys_blocklist=keys_blocklist)

    notice = notifier.build_notice("hello")
    notice["params"] = dict(key1="value1", key2="value2", key3=dict(key1="value1"))
    notice = notifier.send_notice_sync(notice)

    assert notice["params"] == {
        "key1": "[Filtered]",
        "key2": "value2",
        "key3": {"key1": "[Filtered]"},
    }


def _test_deprecated_filter_keys(keys_blacklist):
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        notifier = Notifier(keys_blacklist=keys_blacklist)
        assert len(w) == 1
        assert issubclass(w[-1].category, DeprecationWarning)
        deprecation_message = "keys_blacklist is a deprecated option. "\
                              "Use keys_blocklist instead."
        assert deprecation_message in str(w[-1].message)

    notice = notifier.build_notice("hello")
    notice["params"] = dict(key1="value1", key2="value2", key3=dict(key1="value1"))
    notice = notifier.send_notice_sync(notice)

    assert notice["params"] == {
        "key1": "[Filtered]",
        "key2": "value2",
        "key3": {"key1": "[Filtered]"},
    }


def test_keys_blocklist_exact():
    _test_keys_blocklist(["key1"])
    _test_deprecated_filter_keys(["key1"])


def test_keys_blocklist_regexp():
    _test_keys_blocklist([re.compile("key1")])
    _test_deprecated_filter_keys([re.compile("key1")])


def _test_full_queue():
    notifier = Notifier(max_queue_size=10)

    for _ in range(100):
        future = notifier.notify("hello")

    notifier.close()

    notice = future.result()
    assert notice["error"] == "queue is full"


def _test_rate_limited():
    notifier = Notifier()

    for _ in range(101):
        future = notifier.notify("hello")

    notice = future.result()
    assert notice["error"] == "IP is rate limited"


def test_clean_filename():
    notifier = Notifier()

    filename = notifier._clean_filename("home/lib/python3.6/site-packages/python.py")
    assert filename == "/SITE_PACKAGES/python.py"

def test_error_notifications_disabled():
    notifier = Notifier()
    notifier.config["error_notifications"] = False

    future = notifier.notify("hello")
    notifier.close()

    notice = future.result()
    assert notice["error"] == "error notifications are disabled"



import json
import threading
import time

from http.server import BaseHTTPRequestHandler, HTTPServer

from .test_celery import raise_error

notice = None


class Handler(BaseHTTPRequestHandler):
    def do_POST(self):
        content_length = int(self.headers["Content-Length"])
        post_data = self.rfile.read(content_length).decode("utf-8")

        global notice  # pylint: disable=global-statement
        notice = json.loads(post_data)

        self.send_response(200)
        self.wfile.write('{"id":"1"}'.encode("utf-8"))


def test_celery_integration():
    server_address = ("", 8080)
    server = HTTPServer(server_address, Handler)

    httpd_thread = threading.Thread(target=server.serve_forever)
    httpd_thread.setDaemon(True)
    httpd_thread.start()

    raise_error.apply()

    for _ in range(10):
        if notice is None:
            time.sleep(1)
        else:
            break

    errors = notice["errors"]
    assert len(errors) == 1

    error = errors[0]
    assert error["type"] == "ValueError"
    assert error["message"] == "Test"

    backtrace = error["backtrace"]
    assert len(backtrace) == 2

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_celery.py"
    assert frame["function"] == "raise_error"
    assert frame["line"] == 16
    server.socket.close()



from .notifier import Notifier
from .logging import LoggingHandler, _LOG_RECORD_ATTRS_TO_COPY
from .test_helper import build_logging_record_error, build_logging_record_exception


def test_logging_handler_record_error():
    record = build_logging_record_error("hello", extra=dict(foo="bar"))

    h = LoggingHandler(notifier=Notifier())
    notice = h.build_notice(record)

    errors = notice["errors"]
    assert len(errors) == 1

    error = errors[0]
    assert error["type"] == "test"
    assert error["message"] == "hello"

    backtrace = error["backtrace"]
    assert len(backtrace) == 1

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_helper.py"
    assert frame["function"] == "build_logging_record_error"
    assert frame["line"] == 17
    assert frame["code"] == {
        15: "def build_logging_record_error(*args, **kwargs):",
        16: "    logger, dh = logger_dummy_handler()",
        17: "    logger.error(*args, **kwargs)",
        18: "    return dh.record",
        19: "",
    }

    ctx = notice["context"]
    assert ctx["severity"] == "ERROR"
    assert ctx["component"] == "test_helper"

    params = notice["params"]
    assert params["extra"] == {"foo": "bar"}
    for attr in _LOG_RECORD_ATTRS_TO_COPY:
        assert attr in params


def test_logging_handler_record_exception():
    record = build_logging_record_exception()

    h = LoggingHandler(notifier=Notifier())
    notice = h.build_notice(record)

    errors = notice["errors"]
    assert len(errors) == 1

    error = errors[0]
    assert error["type"] == "ValueError"
    assert error["message"] == "hello"

    backtrace = error["backtrace"]
    assert len(backtrace) == 1

    frame = backtrace[0]
    assert frame["file"] == "/PROJECT_ROOT/pybrake/test_helper.py"
    assert frame["function"] == "build_logging_record_exception"
    assert frame["line"] == 24
    assert frame["code"] == {
        22: "    logger, dh = logger_dummy_handler()",
        23: "    try:",
        24: '        raise ValueError("hello")',
        25: "    except ValueError as err:",
        26: "        logger.exception(err)",
    }



import inspect

from aiohttp import web

from .notifier import Notifier


def create_airbrake_middleware(overrides=None):
    async def factory(app, handler):
        async def middleware(request):
            try:
                response = await handler(request)
                override = overrides.get(response.status)
                if override:
                    return await override(request)
                return response

            except web.HTTPException as ex:
                override = overrides.get(ex.status)
                if override:
                    return await override(request)
                raise

            except Exception as ex:  # pylint: disable=broad-except
                handle_exception(app, ex, request)
                override = overrides.get(500)
                return await override(request)

        return middleware

    return factory


def handle_exception(app, ex, request):
    init_pybrake(app)
    airbrake_send_notification(ex, app["pybrake"], request)


def init_pybrake(app):
    if "pybrake" not in app:
        app["pybrake"] = Notifier(**app["airbrake_config"])


def airbrake_send_notification(ex, notifier, request):
    notice = notifier.build_notice(ex)
    notice["context"].update(additional_context(request))
    notice["params"].update(get_headers(request))
    notifier.send_notice(notice)


def additional_context(request):
    return dict(
        userAgent=get_user_agent(request),
        userAddr=attr_from_request(request, "remote"),
        httpMethod=attr_from_request(request, "method"),
        url=attr_from_request(request, "url"),
    )


def get_user_agent(request):
    headers = get_headers(request)
    return headers.get("User-Agent") if headers else None


def get_headers(request):
    headers = attr_from_request(request, "headers")
    return dict(headers=headers) if headers else {}


def attr_from_request(request, attr_name):
    if hasattr(request, attr_name):
        attr = getattr(request, attr_name)
        return attr() if inspect.ismethod(attr) else attr
    return None



import functools

import celery.exceptions as excs
import celery.app.trace as trace

from . import metrics
from .queues import QueueMetric


_CELERY_EXCEPTIONS = (excs.Retry, excs.Ignore, excs.Reject)


def patch_celery(notifier):
    old_build_tracer = trace.build_tracer

    def build_tracer(name, task, *args, **kwargs):
        if getattr(task, "_ab_patched", False):
            return old_build_tracer(name, task, *args, **kwargs)

        task._ab_patched = True
        task.__call__ = _wrap_task_call(task, task.__call__, notifier=notifier)
        task.run = _wrap_task_call(task, task.run, notifier=notifier)

        return old_build_tracer(name, task, *args, **kwargs)

    trace.build_tracer = build_tracer


def _wrap_task_call(task, fn, *, notifier=None):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        metric = QueueMetric(queue=task.name)
        metrics.set_active(metric)

        try:
            return fn(*args, **kwargs)
        except Exception as exc:
            if not isinstance(exc, _CELERY_EXCEPTIONS):
                notifier.notify(exc)
            raise exc
        finally:
            notifier.queues.notify(metric)
            metrics.set_active(None)

    return wrapper



from functools import lru_cache

from .utils import logger


try:
    from django.conf import settings

    _DJANGO = True
except ImportError:
    _DJANGO = False


@lru_cache()
def get_global_notifier():
    if not _DJANGO:
        return None

    if not hasattr(settings, "AIRBRAKE"):
        if settings.DEBUG:
            logger.info("pybrake is not configured - set settings.AIRBRAKE")
            return None

    from .notifier import Notifier  # pylint: disable=import-outside-toplevel

    return Notifier(**settings.AIRBRAKE)



import os
from functools import lru_cache

from .utils import logger


@lru_cache(maxsize=1000)
def get_git_revision(dirpath):
    try:
        return _get_git_revision(dirpath)
    except (OSError, IOError) as err:
        logger.error("get_git_revision failed: %s", err)
        return None


def _get_git_revision(dirpath):
    head = get_git_head(dirpath)

    prefix = "ref: "
    if not head.startswith(prefix):
        return head
    head = head[len(prefix) :]

    ref_file = os.path.join(dirpath, ".git", head)
    try:
        with open(ref_file, "r", encoding="utf8") as f:
            return f.read().rstrip()
    except (OSError, IOError):
        pass

    refs_file = os.path.join(dirpath, ".git", "packed-refs")
    with open(refs_file, "r", encoding="utf8") as f:
        for line in f:
            if not line or line[0] in ("#", "^"):
                continue

            parts = line.rstrip().split(" ")
            if len(parts) != 2:
                continue

            if parts[1] == head:
                return parts[0]

    return None


def get_git_head(dirpath):
    head_file = os.path.join(dirpath, ".git", "HEAD")
    with open(head_file, "r", encoding="utf8") as f:
        return f.read().rstrip()


def find_git_dir(directory):
    """Returns first directory containing .git file checking the dir and parent dirs."""
    directory = os.path.abspath(directory)
    if not os.path.exists(directory):
        return ""

    for _ in range(10):
        path = os.path.join(directory, ".git")
        if os.path.exists(path):
            return directory

        if directory == "/":
            return ""

        directory = os.path.abspath(os.path.join(directory, os.pardir))

    return ""



import threading
from contextlib import contextmanager
import time as pytime

from .utils import logger


threadLocal = threading.local()

FLUSH_PERIOD = 15
_METRIC_KEY = "_ab_metric"


@contextmanager
def activated_metric(metric):
    set_active(metric)
    yield
    set_active(None)


def set_active(metric):
    setattr(threadLocal, _METRIC_KEY, metric)


def get_active():
    return getattr(threadLocal, _METRIC_KEY, None)


def start_span(name, **kwargs):
    metric = get_active()
    if metric is not None:
        metric.start_span(name, **kwargs)


def end_span(name, **kwargs):
    metric = get_active()
    if metric is not None:
        metric.end_span(name, **kwargs)


class Metric:
    def __init__(self):
        self.start_time = pytime.time()
        self.end_time = None

        self._spans = {}
        self._curr_span = None

        self._groups = {}

    def end(self):
        if self.end_time is None:
            self.end_time = pytime.time()

    def _new_span(self, name, **kwargs):
        return Span(metric=self, name=name, **kwargs)

    def start_span(self, name, *, start_time=None):
        if self._curr_span is not None:
            if self._curr_span.name == name:
                self._curr_span._level += 1
                return
            self._curr_span._pause()

        span = self._spans.get(name)
        if span is None:
            span = self._new_span(name, start_time=start_time)
            self._spans[name] = span
        else:
            span._resume()

        span._parent = self._curr_span
        self._curr_span = span

    def end_span(self, name, *, end_time=None):
        if self._curr_span is not None and self._curr_span.name == name:
            if self._end_span(self._curr_span):
                self._curr_span = self._curr_span._parent
                if self._curr_span is not None:
                    self._curr_span._resume()
            return

        span = self._spans.get(name)
        if span is None:
            logger.error("pybrake: span=%s does not exist", name)
            return
        self._end_span(span, end_time=end_time)

    def _end_span(self, span, *, end_time=None):
        if span._level > 0:
            span._level -= 1
            return False

        span.end(end_time=end_time)
        self._spans.pop(span.name)
        return True

    def _inc_group(self, name, ms):
        self._groups[name] = self._groups.get(name, 0) + ms


class Span:
    def __init__(self, *, metric=None, name="", start_time=None):
        self._metric = metric
        self._parent = None

        self.name = name
        if start_time is not None:
            self.start_time = start_time
        else:
            self.start_time = pytime.time()
        self.end_time = None

        self._dur = 0
        self._level = 0

    def end(self, end_time=None):
        if end_time is not None:
            self.end_time = end_time
        else:
            self.end_time = pytime.time()
        self._dur += (self.end_time - self.start_time) * 1000
        self._metric._inc_group(self.name, self._dur)
        self._metric = None

    def _pause(self):
        if self._paused():
            return
        self._dur += (pytime.time() - self.start_time) * 1000
        self.start_time = 0

    def _paused(self):
        return self.start_time == 0

    def _resume(self):
        if not self._paused():
            return
        self.start_time = pytime.time()



import platform
import urllib.request
import json
import threading
import time

from urllib.parse import urlencode
from copy import deepcopy
from .version import version
from .notifier_name import notifier_name
from .settings_data import SettingsData


# Query params to be appended to each GET request.
_NOTIFIER_INFO = {
    "notifier_name": notifier_name,
    "notifier_version": version,
    "os": platform.platform(),
    "language": platform.python_version(),
}


class RemoteSettings:
    def __init__(self, project_id, host, config):
        self._project_id = project_id
        self._host = host
        self._config = config
        self._data = SettingsData(project_id, {})
        self._prev_data = None

        self._orig_error_notifications = config["error_notifications"]
        self._orig_performance_stats = config["performance_stats"]

    def poll(self):
        thread = threading.Thread(target=self._run)
        thread.daemon = True
        thread.start()

    def _run(self, **kwargs):
        while True:
            try:
                resp = urllib.request.urlopen(self._poll_url(self._data))
                self._prev_data = deepcopy(self._data)
            except urllib.error.HTTPError as err:
                try:
                    if self._prev_data is None:
                        raise err
                    resp = urllib.request.urlopen(self._poll_url(self._prev_data))
                except urllib.error.HTTPError:
                    time.sleep(self._data.interval())
                    continue

            json_data = json.loads(resp.read().decode('utf-8'))
            self._data.merge(json_data)

            error_host = self._data.error_host()
            if error_host is not None:
                self._config["error_host"] = self._data.error_host()

            apm_host = self._data.apm_host()
            if apm_host is not None:
                self._config["apm_host"] = self._data.apm_host()

            self._process_error_notifications(self._data)
            self._process_performance_stats(self._data)

            time.sleep(self._data.interval())

    def _poll_url(self, data):
        url = data.config_route(self._host)
        return url + '?' + urlencode(_NOTIFIER_INFO)

    def _process_error_notifications(self, data):
        if not self._orig_error_notifications:
            return

        self._config["error_notifications"] = data.error_notifications()


    def _process_performance_stats(self, data):
        if not self._orig_performance_stats:
            return

        self._config["performance_stats"] = data.performance_stats()



from .settings_data import SettingsData


def test_interval_when_no_poll_sec():
    s = SettingsData(1, {})
    assert s.interval() == 600

def test_interval_when_poll_sec_is_None():
    s = SettingsData(1, {"poll_sec": None})
    assert s.interval() == 600

def test_interval_when_poll_sec_is_less_than_zero():
    s = SettingsData(1, {"poll_sec": -2})
    assert s.interval() == 600

def test_interval_when_poll_sec_is_zero():
    s = SettingsData(1, {"poll_sec": 0})
    assert s.interval() == 600

def test_interval_when_poll_sec_is_greater_than_zero():
    s = SettingsData(1, {"poll_sec": 123})
    assert s.interval() == 123

def test_config_route_when_missing():
    s = SettingsData(1, {})
    assert s.config_route("http://example.com") == (
        "http://example.com/2020-06-18/config/1/config.json"
    )

def test_config_route_when_it_is_specified():
    s = SettingsData(1, {"config_route": "123/cfg/321/cfg.json"})
    assert s.config_route("http://example.com") == (
        "http://example.com/123/cfg/321/cfg.json"
    )

def test_config_route_when_the_given_host_ends_with_a_trailing_slash():
    s = SettingsData(1, {})
    assert s.config_route("http://example.com/") == (
        "http://example.com/2020-06-18/config/1/config.json"
    )

def test_config_route_when_it_is_None():
    s = SettingsData(1, {"config_route": None})
    assert s.config_route("http://example.com") == (
        "http://example.com/2020-06-18/config/1/config.json"
    )

def test_config_route_when_it_is_empty():
    s = SettingsData(1, {"config_route": ""})
    assert s.config_route("http://example.com") == (
        "http://example.com/2020-06-18/config/1/config.json"
    )

def test_error_notifications_when_it_is_present_and_enabled():
    s = SettingsData(1, {
        "settings": [{
            "name": "errors",
            "enabled": True,
        }],
    })
    assert s.error_notifications()

def test_error_notifications_when_it_is_present_and_disabled():
    s = SettingsData(1, {
        "settings": [{
            "name": "errors",
            "enabled": False,
        }],
    })
    assert not s.error_notifications()

def test_error_notifications_when_it_is_missing():
    s = SettingsData(1, {
        "settings": [],
    })
    assert s.error_notifications()

def test_error_notifications_when_settings_are_missing():
    s = SettingsData(1, {})
    assert s.error_notifications()

def test_performance_stats_when_it_is_present_and_enabled():
    s = SettingsData(1, {
        "settings": [{
            "name": "apm",
            "enabled": True,
        }],
    })
    assert s.performance_stats()

def test_performance_stats_when_it_is_present_and_disabled():
    s = SettingsData(1, {
        "settings": [{
            "name": "apm",
            "enabled": False,
        }],
    })
    assert not s.performance_stats()

def test_performance_stats_when_it_is_missing():
    s = SettingsData(1, {
        "settings": [],
    })
    assert s.performance_stats()

def test_performance_stats_when_settings_are_missing():
    s = SettingsData(1, {})
    assert s.performance_stats()

def test_error_host_when_the_errors_setting_is_present():
    s = SettingsData(1, {
        "settings": [{
            "name": "errors",
            "enabled": True,
            "endpoint": "http://api.example.com",
        }],
    })
    assert s.error_host() == "http://api.example.com"

def test_error_host_when_the_errors_setting_is_present_without_endpoint():
    s = SettingsData(1, {
        "settings": [{
            "name": "errors",
            "enabled": True,
        }],
    })
    assert s.error_host() is None

def test_error_host_when_the_errors_setting_is_missing():
    s = SettingsData(1, {
        "settings": [],
    })
    assert s.error_host() is None

def test_apm_host_when_the_apm_setting_is_present():
    s = SettingsData(1, {
        "settings": [{
            "name": "apm",
            "enabled": True,
            "endpoint": "http://api.example.com",
        }],
    })
    assert s.apm_host() == "http://api.example.com"

def test_apm_host_when_the_apm_setting_is_present_without_endpoint():
    s = SettingsData(1, {
        "settings": [{
            "name": "apm",
            "enabled": True,
        }],
    })
    assert s.apm_host() is None

def test_apm_host_when_the_apm_setting_is_missing():
    s = SettingsData(1, {
        "settings": [],
    })
    assert s.apm_host() is None

def test_merge():
    s = SettingsData(1, {
        "settings": [{"foo": 1}],
    })
    s.merge({"settings": [{"foo": 2}]})
    assert s._data, {
        "settings": [{"foo": 2}]
    }



import json
import collections


_MAX_JSON_SIZE = 64000  # 64kb
_NOTICE_KEYS = ["context", "params", "session", "environment"]


def jsonify_notice(notice):
    b = _jsonify(notice)
    if len(b) < _MAX_JSON_SIZE:
        return b.encode("utf8")

    if "errors" in notice:
        t = Truncator(0)
        for i, error in enumerate(notice["errors"]):
            notice["errors"][i] = t.truncate_dict(error)

    for level in range(10):
        b = _jsonify(notice)
        if len(b) < _MAX_JSON_SIZE:
            break

        t = Truncator(level)
        for key in _NOTICE_KEYS:
            if key in notice:
                notice[key] = t.truncate_dict(notice[key])

    return b.encode("utf8")


def _jsonify(obj):
    return json.dumps(obj, default=_set_default)


def _set_default(obj):
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, (bytes, bytearray)):
        return obj.decode("utf8", "replace")
    if isinstance(
        obj, (collections.UserDict, collections.UserList, collections.UserString)
    ):
        return obj.data
    return str(obj)


class Truncator:
    def __init__(self, level=0):
        self._max_dict_len = _scale(128, level)
        self._max_list_len = _scale(128, level)
        self._max_str_len = _scale(1024, level)

    def truncate_dict(self, d):
        res = {}
        i = 0
        for k, v in d.items():
            res[k] = self.truncate(v)
            i += 1
            if i >= self._max_dict_len:
                break
        return res

    def truncate_list(self, l):
        res = []
        for i, v in enumerate(l):
            res.append(self.truncate(v))
            if i >= self._max_list_len:
                break
        return res

    def truncate_str(self, s):
        if len(s) <= self._max_str_len:
            return s
        return s[: self._max_str_len]

    def truncate(self, v):
        if isinstance(v, str):
            return self.truncate_str(v)
        if isinstance(v, (bytes, bytearray)):
            return self.truncate_str(v.decode("utf8", "replace"))
        if isinstance(v, collections.UserString):
            return self.truncate_str(v.data)

        if isinstance(v, collections.abc.Mapping):
            return self.truncate_dict(v)

        if isinstance(v, collections.abc.Iterable):
            return self.truncate_list(v)

        return v


def _scale(num, level):
    return (num >> level) or 1



import os
import re
import sys
import platform
import socket
import urllib.request
import urllib.error
from concurrent import futures
import json
import time
import warnings

from .notice import jsonify_notice
from .git import find_git_dir
from .routes import _Routes
from .queries import QueryStats
from .queues import QueueStats
from .blocklist_filter import make_blocklist_filter
from .code_hunks import get_code_hunk
from .git import get_git_revision
from .utils import logger
from .version import version
from .notifier_name import notifier_name
from .remote_settings import RemoteSettings


_ERR_IP_RATE_LIMITED = "IP is rate limited"

_AB_URL_FORMAT = "{}/api/v3/projects/{}/notices"

_CONTEXT = dict(
    notifier=dict(
        name=notifier_name, version=version, url="https://github.com/airbrake/pybrake"
    ),
    os=platform.platform(),
    language=f"Python/{platform.python_version()}",
    hostname=socket.gethostname(),
    versions=dict(python=platform.python_version()),
)


class Notifier:
    def __init__(
        self, *, project_id=0, project_key="", host="https://api.airbrake.io", **kwargs
    ):
        self.config = {
            "error_notifications": True,
            "performance_stats": kwargs.get("performance_stats", True),
            "query_stats": kwargs.get("query_stats", True),
            "queue_stats": kwargs.get("queue_stats", True),
            "error_host": host,
            "apm_host": host,
        }
        kwargs["config"] = self.config

        self.routes = _Routes(
            project_id=project_id, project_key=project_key, **kwargs
        )
        self.queries = QueryStats(
            project_id=project_id, project_key=project_key, host=host, **kwargs
        )
        self.queues = QueueStats(
            project_id=project_id, project_key=project_key, host=host, **kwargs
        )

        self._filters = []
        self._rate_limit_reset = 0
        self._max_queue_size = kwargs.get("max_queue_size", 1000)
        self._thread_pool = None
        self._max_workers = kwargs.get('max_workers')

        self._ab_url = _AB_URL_FORMAT.format(host, project_id)
        self._ab_headers = {
            "authorization": "Bearer " + project_key,
            "content-type": "application/json",
        }

        self._context = _CONTEXT.copy()
        self._context["rootDirectory"] = kwargs.get("root_directory", os.getcwd())

        rev = kwargs.get("revision")
        if rev is None:
            # https://devcenter.heroku.com/changelog-items/630
            rev = os.environ.get("SOURCE_VERSION")
        if rev is None:
            git_dir = find_git_dir(self._context["rootDirectory"])
            if git_dir != "":
                rev = get_git_revision(git_dir)

        if rev is not None:
            self._context["revision"] = rev

        if "environment" in kwargs:
            self._context["environment"] = kwargs["environment"]

        self.add_filter(pybrake_error_filter)

        keys_blacklist = kwargs.get("keys_blacklist")
        keys_blocklist = kwargs.get("keys_blocklist")

        if keys_blacklist is not None:
            keys_blocklist = keys_blacklist
            warnings.warn(
                    "keys_blacklist is a deprecated option. "
                    "Use keys_blocklist instead.",
                    DeprecationWarning
                )

        if keys_blocklist is None:
            keys_blocklist = [re.compile("password"), re.compile("secret")]

        self.add_filter(make_blocklist_filter(keys_blocklist))

        if "filter" in kwargs:
            self.add_filter(kwargs["filter"])

        if kwargs.get("remote_config"):
            RemoteSettings(
                project_id,
                "https://notifier-configs.airbrake.io",
                self.config,
            ).poll()

    def close(self):
        if self._thread_pool is not None:
            self._thread_pool.shutdown()

    def add_filter(self, filter_fn):
        """Appends filter to the list.

        Filter is a function that accepts notice. Filter can modify passed
        notice or return None if notice must be ignored.
        """
        self._filters.append(filter_fn)

    def notify_sync(self, err):
        """Notifies Airbrake about exception.

        Under the hood notify is a shortcut for build_notice and send_notice.
        """
        notice = self.build_notice(err)
        return self.send_notice_sync(notice)

    def build_notice(self, err):
        """Builds Airbrake notice from the exception."""
        notice = dict(
            errors=self._build_errors(err),
            context=self._build_context(),
            params=dict(sys_executable=sys.executable, sys_path=sys.path),
        )
        return notice

    def _filter_notice(self, notice):
        for fn in self._filters[::-1]:
            r = fn(notice)
            if r is None:
                notice["error"] = "notice is filtered out"
                return notice, False
            notice = r

        return notice, True

    def send_notice_sync(self, notice):
        """Sends notice to Airbrake.

        It returns notice with 2 possible new keys:
        - {'id' => str} - notice id on success.
        - {'error' => str|Exception} - error on failure.
        """
        notice, ok = self._filter_notice(notice)
        if not ok:
            return notice

        if not self.config.get("error_notifications"):
            return notice

        return self._send_notice_sync(notice)

    def _send_notice_sync(self, notice):
        if time.time() < self._rate_limit_reset:
            notice["error"] = _ERR_IP_RATE_LIMITED
            return notice

        data = jsonify_notice(notice)
        req = urllib.request.Request(self._ab_url, data=data, headers=self._ab_headers)

        try:
            resp = urllib.request.urlopen(req, timeout=5)
        except urllib.error.HTTPError as err:
            resp = err
        except Exception as err:  # pylint: disable=broad-except
            notice["error"] = err
            logger.error(notice["error"])
            return notice

        try:
            body = resp.read()
        except IOError as err:
            notice["error"] = err
            logger.error(notice["error"])
            return notice

        if not (200 <= resp.code < 300 or 400 <= resp.code < 500):
            notice["error"] = f"airbrake: unexpected response status_code={resp.code}"
            notice["error_info"] = dict(code=resp.code, body=body)
            logger.error(notice["error"])
            return notice

        if resp.code == 429:
            return self._rate_limited(notice, resp)

        try:
            body = body.decode("utf-8")
        except UnicodeDecodeError as err:
            notice["error"] = err
            logger.error(notice["error"])
            return notice

        try:
            data = json.loads(body)
        except ValueError as err:  # json.JSONDecodeError requires Python 3.5+
            notice["error"] = err
            logger.error(notice["error"])
            return notice

        if "id" in data:
            notice["id"] = data["id"]
            return notice

        if "message" in data:
            notice["error"] = data["message"]
            logger.error(notice["error"])
            return notice

        notice["error"] = "unexpected Airbrake response"
        notice["error_info"] = dict(data=data)
        logger.error(notice["error"])
        return notice

    def _rate_limited(self, notice, resp):
        v = resp.headers.get("X-RateLimit-Delay")
        if v is None:
            notice["error"] = "X-RateLimit-Delay header is missing"
            return notice

        try:
            delay = int(v)
        except ValueError as err:
            notice["error"] = err
            return notice

        self._rate_limit_reset = time.time() + delay

        notice["error"] = _ERR_IP_RATE_LIMITED
        return notice

    def notify(self, err):
        """Asynchronously notifies Airbrake about exception from separate thread.

        Returns concurrent.futures.Future.
        """
        notice = self.build_notice(err)
        return self.send_notice(notice)

    def send_notice(self, notice):
        """Asynchronously sends notice to Airbrake from separate thread.

        Returns concurrent.futures.Future.
        """
        if not self.config.get("error_notifications"):
            notice["error"] = "error notifications are disabled"
            f = futures.Future()
            f.set_result(notice)
            return f

        notice, ok = self._filter_notice(notice)
        if not ok:
            f = futures.Future()
            f.set_result(notice)
            return f

        pool = self._get_thread_pool()
        if pool._work_queue.qsize() >= self._max_queue_size:
            notice["error"] = "queue is full"
            f = futures.Future()
            f.set_result(notice)
            return f

        return pool.submit(self._send_notice_sync, notice)

    def _build_errors(self, err):
        if err is None:
            return []

        if isinstance(err, str):
            frame = sys._getframe()
            while frame is not None:
                if frame.f_code.co_filename.endswith("pybrake/notifier.py"):
                    frame = frame.f_back
                else:
                    break

            backtrace = self._build_backtrace_frame(frame)
            return [{"message": err, "backtrace": backtrace}]

        errors = []
        while err is not None:
            errors.append(self._build_error(err))
            err = err.__cause__ or err.__context__
        return errors

    def _build_error(self, err):
        backtrace = self._build_backtrace_tb(err.__traceback__)
        error = dict(type=err.__class__.__name__, message=str(err), backtrace=backtrace)
        return error

    def _build_backtrace_tb(self, tb):
        backtrace = []
        for frame, lineno in self._walk_tb(tb):
            f = self._build_frame(frame, lineno)
            if f:
                backtrace.insert(0, f)
        return backtrace

    def _walk_tb(self, tb):
        while tb is not None:
            yield tb.tb_frame, tb.tb_lineno
            tb = tb.tb_next

    def _build_backtrace_frame(self, f):
        backtrace = []
        for frame, lineno in self._walk_stack(f):
            f = self._build_frame(frame, lineno)
            if f:
                backtrace.append(f)
        return backtrace

    def _walk_stack(self, f):
        while f is not None:
            yield f, f.f_lineno
            f = f.f_back

    def _build_frame(self, f, line):
        if f.f_locals.get("__traceback_hide__"):
            return None

        filename = f.f_code.co_filename
        func = f.f_code.co_name

        loader = f.f_globals.get("__loader__")
        module_name = f.f_globals.get("__name__")
        return self._frame_with_code(
            filename, func, line, loader=loader, module_name=module_name
        )

    def _frame_with_code(
        self, filename, func, line, loader=None, module_name=None
    ):  # pylint: disable=too-many-arguments
        frame = dict(file=self._clean_filename(filename), function=func, line=line)

        lines = get_code_hunk(filename, line, loader=loader, module_name=module_name)
        if lines is not None:
            frame["code"] = lines

        return frame

    def _clean_filename(self, s):
        if "/lib/python" in s and "/site-packages/" in s:
            needed = "/site-packages/"
            ind = s.find(needed)
            if ind > -1:
                s = "/SITE_PACKAGES/" + s[ind + len(needed) :]
                return s

        s = s.replace(self._context["rootDirectory"], "/PROJECT_ROOT")
        return s

    def _build_context(self):
        ctx = self._context.copy()

        versions = ctx["versions"]
        for name, mod in sys.modules.copy().items():
            if name.startswith("_"):
                continue
            if hasattr(mod, "__version__"):
                versions[name] = mod.__version__

        return ctx

    def _get_thread_pool(self):
        if self._thread_pool is None:
            if self._max_workers is None:
                self._max_workers = (os.cpu_count() or 1) * 5
            self._thread_pool = futures.ThreadPoolExecutor(max_workers=self._max_workers)
        return self._thread_pool


def pybrake_error_filter(notice):
    backtrace = notice["errors"][0]["backtrace"]
    for frame in backtrace:
        dirname = os.path.dirname(frame["file"])
        file_name = os.path.basename(dirname)
        if file_name == "pybrake":
            if os.path.basename(frame["file"]).startswith("test_"):
                break

            return None

    return notice



import collections
import pytest

from .notice import jsonify_notice


def make_defaultdict():
    d = collections.defaultdict()
    for i in range(100000):
        d[i] = i
    return d


@pytest.mark.parametrize(
    "param, wanted_size",
    [
        (list(range(100000)), 558),
        (set(range(100000)), 558),
        ({i: i for i in range(100000)}, 1339),
        ("x" * 100000, 1049),
        (b"x" * 100000, 1049),
        (bytearray([64] * 100000), 1049),
        (bytearray([64] * 100000 + [128]), 1049),  # UnicodeDecodeError
        (make_defaultdict(), 1339),
        (collections.OrderedDict({i: i for i in range(100000)}), 1339),
        (collections.UserDict({i: i for i in range(100000)}), 1339),
        (collections.UserList(range(100000)), 558),
        (collections.UserString("x" * 100000), 1049),
        (collections.UserString(b"x" * 100000), 1049),
    ],
)
def test_truncation(param, wanted_size):
    notice = dict(params=dict(param=param))

    b = jsonify_notice(notice)
    assert len(b) == wanted_size



import logging


def get_exception():
    try:
        raise ValueError("hello")
    except ValueError as err:
        return err


def build_notice_from_str(notifier, s):
    return notifier.build_notice(s)


def build_logging_record_error(*args, **kwargs):
    logger, dh = logger_dummy_handler()
    logger.error(*args, **kwargs)
    return dh.record


def build_logging_record_exception():
    logger, dh = logger_dummy_handler()
    try:
        raise ValueError("hello")
    except ValueError as err:
        logger.exception(err)
        return dh.record


def logger_dummy_handler():
    class DummyHandler(logging.Handler):
        def __init__(self):
            super().__init__()
            self.record = None

        def emit(self, record):
            self.record = record

    logger = logging.getLogger("test")
    dh = DummyHandler()
    logger.addHandler(dh)

    return logger, dh


def get_nested_exception():
    try:
        raise get_exception()
    except ValueError as err:
        try:
            raise ValueError("world") from err
        except ValueError as err:
            return err


def get_exception_in_cython():
    # accumulation_tree is required by tdigest
    from accumulation_tree import AccumulationTree  # pylint: disable=import-outside-toplevel
    t = AccumulationTree(lambda x: x)
    t.insert(1, '1')
    try:
        return t.insert('1', '1')
    except TypeError as err:
        return err



import os

from .git import find_git_dir


def test_find_git_dir():
    work_dir = os.getcwd()

    tests = [
        {"dir": "", "ok": True},
        {"dir": "./", "ok": True},
        {"dir": "...", "ok": False},
        {"dir": "../pybrake", "ok": True},
        {"dir": work_dir, "ok": True},
        {"dir": os.path.join(work_dir, "pybrake"), "ok": True},
        {"dir": os.path.join(work_dir, "pybrake", "__pycache__"), "ok": True},
        {"dir": "../", "ok": False},
        {"dir": "abc", "ok": False},
        {"dir": os.path.join(work_dir, "abc"), "ok": False},
        {"dir": os.path.join(work_dir, ".."), "ok": False},
    ]

    for test in tests:
        git_dir = find_git_dir(test["dir"])
        if test["ok"]:
            assert git_dir == work_dir
        else:
            assert git_dir == ""



import logging

from .notifier import Notifier
from .global_notifier import get_global_notifier


def _log_record_attrs():
    record = logging.LogRecord("", logging.ERROR, "", 0, "", None, None)
    attrs = set()
    for k in vars(record).keys():
        attrs.add(k)
    attrs.add("message")  # set by Formatter
    return attrs


_LOG_RECORD_ATTRS = _log_record_attrs()


_LOG_RECORD_ATTRS_TO_COPY = [
    "created",
    "msecs",
    "process",
    "processName",
    "relativeCreated",
    "thread",
    "threadName",
]


class LoggingHandler(logging.Handler):
    def __init__(self, notifier=None, level=logging.ERROR, **kwargs):
        logging.Handler.__init__(self, level=level)
        if notifier is None:
            notifier = get_global_notifier() or Notifier(**kwargs)
        self._notifier = notifier

    def emit(self, record):
        try:
            notice = self.build_notice(record)
            self._notifier.send_notice(notice)
        except Exception:  # pylint: disable=broad-except
            self.handleError(record)

    def build_notice(self, record):
        notice = self._notifier.build_notice(None)
        notice["errors"].append(self._build_error(record))
        self._update_context(notice["context"], record)
        self._update_params(notice["params"], record)
        return notice

    def _build_error(self, record):
        if record.exc_info:
            return self._build_error_from_exc_info(record.exc_info)

        error = dict(
            type=record.name,
            message=record.getMessage(),
            backtrace=self._build_backtrace(record),
        )
        return error

    def _build_error_from_exc_info(self, exc_info):
        cls, err, tb = exc_info
        backtrace = self._notifier._build_backtrace_tb(tb)
        error = dict(type=cls.__name__, message=str(err), backtrace=backtrace)
        return error

    def _build_backtrace(self, record):
        frame = self._notifier._frame_with_code(
            record.pathname, record.funcName, record.lineno
        )
        backtrace = [frame]
        return backtrace

    def _update_context(self, context, record):
        context["messagePattern"] = record.msg
        context["severity"] = record.levelname
        context["component"] = record.module

    def _update_params(self, params, record):
        extra = self._build_extra(record)
        if extra is not None:
            params["extra"] = extra

        for attr in _LOG_RECORD_ATTRS_TO_COPY:
            v = getattr(record, attr)
            if v is not None:
                params[attr] = v

    def _build_extra(self, record):
        extra = {}
        for k, v in vars(record).items():
            if k not in _LOG_RECORD_ATTRS:
                if extra is None:
                    extra = {k: v}
                else:
                    extra[k] = v
        return extra



import re
import collections.abc

from .notice import _NOTICE_KEYS


_FILTERED = "[Filtered]"


def make_blocklist_filter(keys_blocklist):
    def blocklist_filter(notice):
        for key in _NOTICE_KEYS:
            if key in notice:
                _filter_dict(notice[key], keys_blocklist)
        return notice

    return blocklist_filter


def _filter_dict(d, keys_blocklist):
    for key, value in d.items():
        if _is_blocklisted(key, keys_blocklist):
            d[key] = _FILTERED
            continue
        if isinstance(value, collections.abc.Mapping):
            _filter_dict(value, keys_blocklist)


def _is_blocklisted(key, keys_blocklist):
    for k in keys_blocklist:
        if k == key:
            return True
        if _is_regexp(k) and k.match(key):
            return True
    return False


try:
    _pattern_type = re._pattern_type
except AttributeError:
    try:
        _pattern_type = re.Pattern
    except AttributeError:
        pass


def _is_regexp(v):
    return isinstance(v, _pattern_type)



import time
import functools
import threading

from django.conf import settings
from django.utils.module_loading import import_string
from django.db import connections
from django.template import Template
from django.core import cache
from django.core.cache import CacheHandler
from django.middleware import cache as middleware_cache

from .global_notifier import get_global_notifier
from .route_metric import RouteMetric
from . import metrics


_UNKNOWN_ROUTE = "UNKNOWN"
_REQUEST_KEY = "_django_request"

threadLocal = threading.local()


def set_request(request):
    setattr(threadLocal, _REQUEST_KEY, request)


def get_request():
    return getattr(threadLocal, _REQUEST_KEY, None)


def request_filter(notice):
    request = getattr(threadLocal, _REQUEST_KEY, None)
    if request is None:
        return notice

    req_filter = get_exception_reporter_filter(request)
    notice["params"]["request"] = dict(
        scheme=request.scheme,
        method=request.method,
        GET=dict(request.GET),
        POST=dict(req_filter.get_post_parameters(request)),
        META=dict(request.META),
        FILES=dict(request.FILES),
        COOKIES=dict(request.COOKIES),
        session=dict(request.session),
    )

    return notice


def template_render(self, context):
    metrics.start_span("template")
    res = self.nodelist.render(context)
    metrics.end_span("template")
    return res


if Template._render != template_render:  # pylint: disable=comparison-with-callable
    Template.original_render = Template._render
    Template._render = template_render


class AirbrakeMiddleware:
    def __init__(self, get_response):
        self._notifier = get_global_notifier()
        self._config = self._notifier.config
        self.get_response = get_response

        self._notifier.add_filter(request_filter)

    def __call__(self, request):
        if not self._config.get("performance_stats"):
            return self.get_response(request)

        set_request(request)

        for conn in connections.all():
            wrap_cursor(conn, self._notifier)

        metric = RouteMetric(method=request.method)
        with metrics.activated_metric(metric):
            response = self.get_response(request)

        metric.status_code = response.status_code
        if "Content-Type" in response:
            metric.content_type = response["Content-Type"]
        metric.end_time = time.time()
        self._notifier.routes.notify(metric)

        return response

    def process_view(self, request, view_func, view_args, view_kwargs):
        metric = metrics.get_active()
        if metric is None:
            return

        if request.resolver_match.view_name:
            metric.route = request.resolver_match.view_name
        else:
            route = view_func.__module__
            route += "." + view_func.__name__
            metric.route = route

        set_request(None)

    def process_exception(self, request, exception):
        if not self._notifier:
            return

        notice = self._notifier.build_notice(exception)

        ctx = notice["context"]
        ctx["url"] = request.build_absolute_uri()
        ctx["route"] = request.resolver_match.url_name

        user_addr = get_remote_addr(request)
        if user_addr:
            ctx["userAddr"] = user_addr

        if request.user.is_authenticated:
            user = request.user
            user_info = dict(username=user.get_username(), name=user.get_full_name())
            if hasattr(user, "get_email_field_name"):
                user_info["email"] = getattr(user, user.get_email_field_name())
            elif hasattr(user, "email"):
                user_info["email"] = user.email
            ctx["user"] = user_info

        self._notifier.send_notice(notice)


def get_remote_addr(request):
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        return x_forwarded_for.split(",")[0]
    return request.META.get("REMOTE_ADDR")


@functools.lru_cache()
def get_default_exception_reporter_filter():
    # Instantiate the default filter for the first time and cache it.
    return import_string(settings.DEFAULT_EXCEPTION_REPORTER_FILTER)()


def get_exception_reporter_filter(request):
    default_filter = get_default_exception_reporter_filter()
    return getattr(request, "exception_reporter_filter", default_filter)


def wrap_cursor(conn, notifier):
    if hasattr(conn, "_ab_cursor"):
        return conn.cursor
    conn._ab_cursor = conn.cursor

    def cursor_wrapper(*args, **kwargs):
        cursor = conn._ab_cursor(*args, **kwargs)
        return CursorWrapper(conn, notifier, cursor)

    conn.cursor = cursor_wrapper
    return cursor_wrapper


class CursorWrapper:
    def __init__(self, conn, notifier, cursor):
        self._notifier = notifier
        self._cursor = cursor

    def callproc(self, procname, params=None):
        return self._record(self._cursor.callproc, procname, params)

    def execute(self, sql, params=None):
        return self._record(self._cursor.execute, sql, params)

    def executemany(self, sql, param_list):
        return self._record(self.cursor.executemany, sql, param_list)

    def _record(self, method, sql, params):
        metric = metrics.get_active()

        start_time = time.time()
        metrics.start_span("sql", start_time=start_time)
        try:
            return method(sql, params)
        finally:
            end_time = time.time()
            metrics.end_span("sql", end_time=end_time)
            if hasattr(sql, "as_string"):
                sql = sql.as_string(self._cursor.cursor)
            self._notifier.queries.notify(
                query=sql,
                method=getattr(metric, "method", ""),
                route=getattr(metric, "route", ""),
                start_time=start_time,
                end_time=end_time,
            )

    def __getattr__(self, attr):
        return getattr(self._cursor, attr)

    def __iter__(self):
        return iter(self._cursor)

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        self._cursor.close()


def cache_span(fn):
    @functools.wraps(fn)
    def wrapped(self, *args, **kwargs):
        metrics.start_span("cache")
        res = fn(self, *args, **kwargs)
        metrics.end_span("cache")
        return res

    return wrapped


class CacheWrapper:
    def __init__(self, actual_cache):
        self.cache = actual_cache

    def __repr__(self):
        return str("<CacheWrapper for %s>") % repr(self.cache)

    def __contains__(self, key):
        return self.cache.__contains__(key)

    def __getattr__(self, name):
        return getattr(self.cache, name)

    @cache_span
    def add(self, *args, **kwargs):
        return self.cache.add(*args, **kwargs)

    @cache_span
    def get(self, *args, **kwargs):
        return self.cache.get(*args, **kwargs)

    @cache_span
    def set(self, *args, **kwargs):
        return self.cache.set(*args, **kwargs)

    @cache_span
    def touch(self, *args, **kwargs):
        return self.cache.touch(*args, **kwargs)

    @cache_span
    def delete(self, *args, **kwargs):
        return self.cache.delete(*args, **kwargs)

    @cache_span
    def clear(self, *args, **kwargs):
        return self.cache.clear(*args, **kwargs)

    @cache_span
    def has_key(self, *args, **kwargs):
        # Ignore flake8 rules for has_key since we need to support caches
        # that may be using has_key.
        return self.cache.has_key(*args, **kwargs)  # noqa

    @cache_span
    def incr(self, *args, **kwargs):
        return self.cache.incr(*args, **kwargs)

    @cache_span
    def decr(self, *args, **kwargs):
        return self.cache.decr(*args, **kwargs)

    @cache_span
    def get_many(self, *args, **kwargs):
        return self.cache.get_many(*args, **kwargs)

    @cache_span
    def set_many(self, *args, **kwargs):
        self.cache.set_many(*args, **kwargs)

    @cache_span
    def delete_many(self, *args, **kwargs):
        self.cache.delete_many(*args, **kwargs)

    @cache_span
    def incr_version(self, *args, **kwargs):
        return self.cache.incr_version(*args, **kwargs)

    @cache_span
    def decr_version(self, *args, **kwargs):
        return self.cache.decr_version(*args, **kwargs)


class AirbrakeCacheHandler(CacheHandler):
    def __getitem__(self, alias):
        actual_cache = super().__getitem__(alias)
        return CacheWrapper(actual_cache)


cache_handler = AirbrakeCacheHandler()
middleware_cache.caches = cache_handler
cache.caches = cache_handler



from .notifier import Notifier
from .logging import LoggingHandler
from .version import version as __version__
from .route_metric import RouteMetric, RouteBreakdowns
from .queries import QueryStat
from .queues import QueueMetric



import logging
from datetime import datetime


def time_trunc_minute(time):
    t = datetime.utcfromtimestamp(time).replace(second=0, microsecond=0)
    return t.strftime("%Y-%m-%dT%H:%M:%SZ")


def _get_logger():
    _logger = logging.getLogger("pybrake")
    fmter = logging.Formatter(
        "%(asctime)s %(filename)s:%(lineno)d %(name)s %(levelname)s - %(message)s"
    )

    sh = logging.StreamHandler()
    sh.setFormatter(fmter)
    _logger.addHandler(sh)

    return _logger


logger = _get_logger()



from functools import lru_cache


@lru_cache(maxsize=1000)
def get_code_hunk(filename, lineno, nlines=2, loader=None, module_name=None):
    lines = _get_lines_from_file(filename, loader=loader, module_name=module_name)
    if lines is None:
        return None

    start = max(0, lineno - 1 - nlines)
    end = min(lineno + nlines, len(lines))
    hunk = {}
    for i, line in enumerate(lines[start:end], start + 1):
        if isinstance(line, bytes):
            line = line.decode("utf8", "replace")
        hunk[i] = line.rstrip("\n")

    return hunk


@lru_cache(maxsize=10)
def _get_lines_from_file(filename, loader=None, module_name=None):
    if loader is not None and hasattr(loader, "get_source"):
        try:
            source = loader.get_source(module_name)
        except ImportError:
            pass
        else:
            if source:
                return source.splitlines()

    try:
        with open(filename, "rb") as f:
            return f.read().splitlines()
    except (OSError, IOError):
        return None
